@@ -3,8 +3,8 @@ use t::util;
 use Carp;
 use Test::More;
 use Test::Deep qw(cmp_details deep_diag subbagof);
-use List::Util qw(min);
-use List::MoreUtils qw(uniq);
+use List::Util qw(first min);
+use List::MoreUtils qw(uniq any);
 use Hash::AutoHash::Args;
 # use Hash::AutoHash::MultiValued;
 use Exporter();
@@ -13,24 +13,32 @@ our @ISA=qw(Exporter);
 
 our @EXPORT=
   (@t::util::EXPORT,
-   qw(check_object_basics sort_objects vary_case
-      prep_tabledata load_maptable load_master load_ur 
-      select_ur filter_ur count_ur select_ur_sanity 
-      check_table check_database_sanity check_maptables_sanity check_masters_sanity 
-      cleanup_db cleanup_ur
-      cmp_objects cmp_objects_quietly cmp_table cmp_table_quietly cmp_table_nocase
-      cmp_op cmp_op_quietly cmp_op_quickly
+   qw(check_object_basics sort_objects power_subsets cross_product vary_case uniq_rows sql_in
+      prep_tabledata load_maptable load_master load_ur select_ur select_ur_sanity cleanup_db
+      check_database_sanity check_maptables_sanity check_masters_sanity 
       check_handcrafted_idtypes check_handcrafted_masters check_handcrafted_maptables
       check_handcrafted_name2idtype check_handcrafted_name2master check_handcrafted_name2maptable
       check_handcrafted_id2object check_handcrafted_id2name check_implicit_masters
       load_handcrafted_maptables load_handcrafted_masters
+      cmp_table cmp_table_quietly cmp_table_nocase
+      cmp_op cmp_op_quietly cmp_op_quickly
+      pnames pgraph
     ));
+# NG 13-07-16: not used now but keep for future
+push(@EXPORT,qw(grep_rows sort_rows check_table 
+		cmp_objects cmp_objects_quietly order_tables));
+# NG 13-07-16: used internally
+#              filter_ur count_ur cleanup_ur
+# NG 13-07-16: obsolete. 
+#              sort_name_lists
 
 sub check_object_basics {
   my($object,$class,$name,$label)=@_;
   report_fail($object,"$label connected object defined") or return 0;
   $object->name;		# touch object in case still Oid
-  report_fail(UNIVERSAL::isa($object,$class),"$label: class") or return 0;
+  # NG 13-09-24: changed to isa_ok_quietly
+  # report_fail(UNIVERSAL::isa($object,$class),"$label: class") or return 0;
+  isa_ok_quietly($object,$class,$label) or return 0;
   report_fail($object->name eq $name,"$label: name") or return 0;
   return 1;
 }
@@ -57,6 +65,72 @@ sub sort_objects {
   my @sorted_objects=sort {$a->name cmp $b->name} @$objects;
   wantarray? @sorted_objects: \@sorted_objects;
 }
+# NG 13-07-07: generate subsets of power set. much faster than Set::Scalar
+#             for large sets if number of subsets not too big
+# 1st argument is either ARRAY ref or integer
+#   if integer, it's universe size; universe is 0..$size-1
+#   if ARRAY, it's universe
+# min, max are sizes of subsets emitted
+# if only one provided, it's max
+# if none provided, it's regular power_set. equivalent to min=0, max=size
+sub power_subsets {
+  my($size,$universe);
+  if (ref($_[0])) {
+    $universe=shift;
+    $size=scalar @$universe;
+  } else {
+    $size=shift;
+  }
+  my($min,$max)=@_==1? (0,@_): (@_>1? @_: (0,$size));
+  my @subsets=(defined $max && $max==0)? []: _power_subsets(0,$size,$min,$max);
+  @subsets=map {[@$universe[@{$_}]]} @subsets if defined $universe;
+  wantarray? @subsets: \@subsets;
+}
+sub _power_subsets {
+  my($i,$size,$min,$max)=@_;
+  return ([],[$i]) if $i==$size-1;
+  my @recurse=_power_subsets($i+1,$size,$min,$max);
+  # cases: size refers to size of downstream subset
+  # 1) size==max. implies size>=min. pass along, don't add current
+  # 2) size<max && size+i>=min.      pass along, and add current
+  # 3) size<max && size+1+i>=min.    add current
+  # 4) else.  do nothing
+  my @out=map {
+    scalar(@$_)==$max? $_ :
+      (scalar(@$_)+$i>=$min? ($_,[$i,@$_]): 
+       (scalar(@$_)+1+$i>=$min? [$i,@$_]: ()))} @recurse;
+
+  @out;
+}
+# NG 13-07-20: cartesion product of a list of vectors (ARRAYs)
+# arguments are either ARRAY refs or integers
+#   if integer, it's universe size; universe is 0..$size-1
+#   if ARRAY, it's universe
+sub cross_product {
+  my @out;
+  if (any {ref $_} @_) {
+    my @sets=@_;
+    my @sizes=map {ref $_? scalar(@$_): $_} @sets;
+    my @idx_rows=_cross_product(@sizes);
+    for my $idx_row (@idx_rows)  {
+      my @row=map
+	{my $idx=$idx_row->[$_]; ref $sets[$_]? $sets[$_]->[$idx]: $idx;} 0..@sets-1;
+      push(@out,\@row);
+    }
+  } else {
+    @out=_cross_product(@_);
+  }
+  wantarray? @out: \@out;
+}
+sub _cross_product {
+  my @sizes=@_;
+  return unless @sizes;
+  my $size=shift @sizes;
+  return map {[$_]} 0..$size-1 unless @sizes;
+  my @recurse=_cross_product(@sizes);
+  map {my $i=$_; map {[$i,@$_]} @recurse} 0..$size-1;
+}
+
 # NG 13-06-15: added vary_case to test case insensitve comparisons
 # vary case for list of ids
 sub vary_case {
@@ -188,15 +262,24 @@ sub load_master {
 # create universal relation (UR)
 # algorithm: natual full outer join of all maptables and explicit masters
 #            any pre-order traversal of schema graph will work (I think!)
-# >>> assume that lexical order of maptables gives a valid pre-order <<<
 # sadly, since MyQSL still lacks full outer joins, have to emulate with left/right
 # joins plus union. do it step-by-step: I couldn't figure out how to do it in
 # one SQL statement...
+# >>> assume that lexical order of maptables gives a valid pre-order <<<
+# NG 13-06-12: assumption above is certainly worng in some tests and probably wrong 
+#              in many others.  scary this wasn't cught sooner.
+#              as a first step towards a solution, add @tables arg to let caller
+#              pass in a valid join order
 sub load_ur {
-  my($babel,$urname)=@_;
+  my($babel,$urname,@tables)=@_;
   $urname or $urname='ur';
-  # ASSUME that lexical order of maptables gives a valid pre-order
-  my @tables=sort {$a->tablename cmp $b->tablename} @{$babel->maptables};
+  # if @tables not given, ASSUME lexical order of maptables is valid pre-order
+  # this will change soon
+  if (@tables) {
+    @tables=map {$babel->name2maptable($_)} @tables;
+  } else {
+    @tables=sort {$a->tablename cmp $b->tablename} @{$babel->maptables};
+  }
   # add in explicit Masters. order doesn't matter so long as they're last
   push(@tables,grep {$_->explicit} @{$babel->masters});
   # %column2type maps column_names to sql types
@@ -222,11 +305,12 @@ sub load_ur {
 # NG 12-11-20: fixed input column for histories: 0th column is '_X_' if input has history
 # NG 12-11-23: added validate
 # NG 13-06-10: changed to do case insensitive comparisons, eg, searching for 'htt' as gene_symbol
+# NG 13-10-14: added query
 # select data from ur (will actually work for any table)
 sub select_ur {
   my $args=new Hash::AutoHash::Args(@_);
-  my($babel,$urname,$input_idtype,$input_ids,$output_idtypes,$filters,$validate,$nocase)=
-    @$args{qw(babel urname input_idtype input_ids output_idtypes filters validate nocase)};
+  my($babel,$urname,$input_idtype,$input_ids,$output_idtypes,$filters,$query,$validate,$nocase)=
+    @$args{qw(babel urname input_idtype input_ids output_idtypes filters query validate nocase)};
   confess "input_idtype must be set. call select_ur_sanity instead" unless $input_idtype;
   # confess "Only one of inputs_ids or input_ids_all may be set" if $input_ids && $input_ids_all;
   $urname or $urname=$args->tablename || 'ur';
@@ -245,7 +329,7 @@ sub select_ur {
   my $dbh=$babel->autodb->dbh;
   # NG 10-08-25: removed 'uniq' since duplicate columns are supposed to be kept
   # my @columns=uniq grep {length($_)} ($input_idtype,@output_idtypes);
-  # NG 12-09-04: include filterut_idtypes so we can do filtering in Perl
+  # NG 12-09-04: include filter_idtypes so we can do filtering in Perl
   # NG 12-09-04: test for length obsolete, since input_idtype required
   # my @columns=grep {length($_)} ($input_idtype,@filter_idtypes,@output_idtypes);
   # NG 12-11-20: 0th column is '_X_' if input has history
@@ -259,6 +343,9 @@ sub select_ur {
   # hang onto valid input ids if doing validate
   # NG 13-06-10: added 'lc' for case insensitive comparisons
   my %valid=map {lc $_->[0]=>1} @$table if $validate;
+  # NG 13-10-14: added query. have to do query after validate.
+  $sql.=" AND ($query)" if $query;
+  my $table=$dbh->selectall_arrayref($sql);
 
   # now do filtering. columns are input, filters, then outputs, finally history columns
   my %name2idx=map {$columns[$_]=>$_} 0..$#columns;
@@ -294,9 +381,48 @@ sub select_ur {
     # NG 13-06-15: added '||0' for case insensitive comparisons
     push(@$table,map {[$_,$id2valid{$_}||0,(undef)x@$output_idtypes]} @missing_ids);
   }
+  # NG 13-07-15: remove partial duplicates
+  $table=remove_pdups($table) unless $args->keep_pdups;
   $table;
 }
 
+sub remove_pdups {
+  my $table=shift;
+
+  my %groups=group {$_->[0]} @$table;
+  my $pseudo_dups=0;
+  $table=[];
+  for my $group (values %groups) {
+    push(@$table,@$group),next unless scalar(@$group)>1;
+    my @rows=sort {undefs($a)<=>undefs($b)} @$group;
+    my %rows=map {$_=>$rows[$_]} (0..$#rows);
+    my %pseudo_dups;
+    for(my $i=0; $i<@rows-1; $i++) {
+      next unless $rows{$i};
+      for(my $j=$i+1; $j<@rows; $j++) {
+	next unless $rows{$j};
+	delete $rows{$j}, $pseudo_dups++ if pseudo_dup($rows[$i],$rows[$j]);
+      }
+    }
+    push(@$table,values %rows);
+  }
+  # TODO: put this under some sort of flag...
+  # diag('+++ select_ur pseudo_dups='.$pseudo_dups) if $pseudo_dups;
+  # diag('+++ select_ur pseudo_dups='.$pseudo_dups);
+
+  $table;
+}
+# used to sort by number of undefs
+sub undefs {scalar grep {!defined $_} @{$_[0]};}
+# row $j is pseudo-dup of $i if they agree wherever both defined, else $j is undef
+sub pseudo_dup {
+  my($rowi,$rowj)=@_;
+  for(my $k=0; $k<=@$rowi; $k++) {
+    return 0 if defined $rowj->[$k] && $rowi->[$k] ne $rowj->[$k];
+  }
+  1;
+}
+
 # NG 13-06-10: changed to do case insensitive comparisons
 sub filter_ur {
   my($table,$col,$ids)=@_;
@@ -329,13 +455,45 @@ sub uniq_rows {
   # NG 13-06-10: added 'lc' for case insensitive comparisons
   my @row_strings=map {lc join($;,@$_)} @$rows;
   my %seen;
-  my $uniq_rows=[];
+  my @uniq_rows;
   for(my $i=0; $i<@$rows; $i++) {
     my $row_string=$row_strings[$i];
-    push(@$uniq_rows,$rows->[$i]) unless $seen{$row_string}++;
+    push(@uniq_rows,$rows->[$i]) unless $seen{$row_string}++;
   }
-  $uniq_rows;
+  # $uniq_rows;
+  wantarray? @uniq_rows: \@uniq_rows;
 }
+# NG 13-06-21: moved from 050/translate.pm and renamed from grep_table
+sub grep_rows {
+  my($table,$col,$ids)=@_;
+  my $pattern=join('|',map {"\^$_\$"} @$ids);
+  $pattern=qr/$pattern/;
+  [grep {$_->[$col]=~/$pattern/} @$table];
+}
+# NG 13-06-19: sort arrays of names, typically output subsets produced by power_set
+# NG 13-07-16: renamed 
+sub sort_rows {
+  sort {@$a<=>@$b || first {$_} map {$a->[$_] cmp $b->[$_]} 0..$#$a} @_;
+}
+# make SQL IN clauses for test queries. 
+# arguments are 1) single value, 2) ARRAY of values, or 3) column=>values pairs
+sub sql_in {
+  my $sql;
+  if (@_==1) {
+    my @values=flatten(shift);
+    $sql='IN ('.join(',',map {"'$_'"} @values).')';
+  } else {
+    my %args=@_;
+    my @sql;
+    while (my($column,$values)=each %args) {
+      my @values=flatten($values);
+      push(@sql,"$column IN (".join(',',map {"'$_'"} @values).')');
+    }
+    $sql=join(' AND ',@sql);
+  }
+  $sql;
+}
+
 # process filters ARRAY - a bit hacky 'cuz filter=>undef not same as filter=>[undef]
 sub filters_array {
   my @filters=@{$_[0]};
@@ -454,8 +612,8 @@ sub check_masters_sanity {
 # 1) reports errors the way we want them
 # 2) sorts the args to avoid Test::Deep's 'bag' which is ridiculously slow...
 sub cmp_objects {
-  my($actual,$correct,$label,$file,$line,$limit)=@_;
-  my $ok=cmp_objects_quietly($actual,$correct,$label,$file,$line,$limit);
+  my($actual,$correct,$label,$file,$line)=@_;
+  my $ok=cmp_objects_quietly($actual,$correct,$label,$file,$line);
   report_pass($ok,$label);
 }
 sub cmp_objects_quietly {
@@ -567,6 +725,32 @@ sub cmp_rows {
   # equal up to here. if $b has more, then $a is smaller
   $#$a <=> $#$b;
 }
+# NG 13-06-12: construct pre-order traversal of schema 'table-graph';
+sub order_tables {
+  my($babel)=@_;
+  my $graph=new Graph::Undirected;
+  my @tables=@{$babel->maptables};
+  for my $table (@tables) {
+    my @idtypes=@{$table->idtypes};
+    my @neighbors=grep {$_!=$table} map {@{$_->maptables}} @idtypes;
+    my $me=$table->name;
+    map {$graph->add_edge($me,$_->name)} @neighbors;
+  }
+  my $tree=$graph->minimum_spanning_tree;
+  # now do a pre-order traversal, eg, depth-first
+  my @nodes=$tree->vertices;
+  my %avbl=map {$_=>1} @nodes;
+  my @future=$nodes[0];
+  my @traversal;
+  while (@future) {
+    my $node=shift @future;
+    next unless $avbl{$node};	# don't process if already visited
+    push(@traversal,$node);
+    $avbl{$node}=0;		# mark as visited
+    push(@future,grep {$avbl{$_}} $tree->neighbors($node));
+  }
+  wantarray? @traversal: \@traversal;
+}
 # emulate natural full outer join. return result table
 # $result is optional name of result table. if not set, unique name generated
 # TODO: add option to delete intermediate tables as we go.
@@ -619,7 +803,7 @@ sub cleanup_db {
 sub cleanup_ur {t::FullOuterJoinTable->cleanup(@_) }
 
 ########################################
-# these functions test our hand-crafted Babel & components
+# these functions test our 'standard' hand-crafted Babel & components
 
 sub check_handcrafted_idtypes {
   my($actual,$mature,$label)=@_;
@@ -634,7 +818,8 @@ sub check_handcrafted_idtypes {
     report_fail(UNIVERSAL::isa($actual,$class),"$label object $i: class") or return 0;
     report_fail($actual->name eq "type_$suffix","$label object $i: name") or return 0;
     report_fail($actual->id eq "idtype:type_$suffix","$label object $i: id") or return 0;
-    report_fail($actual->display_name eq "display_name_$suffix","$label object $i: display_name") or return 0;
+    report_fail($actual->display_name eq "display_name_$suffix",
+		"$label object $i: display_name") or return 0;
     report_fail($actual->referent eq "referent_$suffix","$label object $i: referent") or return 0;
     report_fail($actual->defdb eq "defdb_$suffix","$label object $i: defdb") or return 0;
     report_fail($actual->meta eq "meta_$suffix","$label object $i: meta") or return 0;
@@ -643,9 +828,21 @@ sub check_handcrafted_idtypes {
     report_fail(as_bool($actual->internal)==0,"$label object $i: internal") or return 0;
     report_fail(as_bool($actual->external)==1,"$label object $i: external") or return 0;
     if ($mature) {
-      check_object_basics($actual->babel,'Data::Babel','test',"$label object $i babel");
+      my $babel=$actual->babel;
+      check_object_basics($babel,'Data::Babel','test',"$label object $i babel");
       check_object_basics($actual->master,'Data::Babel::Master',
 			  "type_${suffix}_master","$label object $i master");
+      # NG 13-09-24: added tests for maptables
+      my $maptables=$actual->maptables;
+      map {isa_ok_quietly($_,'Data::Babel::MapTable',"$label object $i maptable")} @{$maptables}
+	or return 0;
+      my @correct_maptables=
+	($suffix eq '001'? map {$babel->name2maptable("maptable_$_")} qw(001):
+	 ($suffix eq '002'? map {$babel->name2maptable("maptable_$_")} qw(001 002): 
+	  ($suffix eq '003'? map {$babel->name2maptable("maptable_$_")} qw(002 003): 
+	   ($suffix eq '004'? map {$babel->name2maptable("maptable_$_")} qw(003):
+	    confess "Unexpected suffix $suffix"))));
+      cmp_bag_quietly($maptables,\@correct_maptables,"$label object $i maptables") or return 0;
     }
   }
   pass($label);
@@ -665,64 +862,21 @@ sub check_handcrafted_masters {
     my $suffix='00'.($i+1);
     my $name="type_${suffix}_master";
     my $id="master:$name";
+    # NG 13-09-02: DEPRECATED workflow related attributes
+    my($query,$view,$implicit);
     # masters 2&3 are implicit, hence some of their content is special
-    my($inputs,$namespace,$query,$view,$implicit);
-    if ($i<2) {
-      $inputs="MainData/table_$suffix";
-      $namespace="ConnectDots";
-      $namespace="ConnectDots";
-      $query="SELECT col_$suffix AS type_$suffix FROM table_$suffix";
-      $view=0;
-      $implicit=0;
-    } else {
-      $namespace='';		# namespace not in input config file, but hopefully set in output
-      $implicit=1;
-      if ($i==2) {
-	$inputs="ConnectDots/maptable_003 ConnectDots/maptable_002";
-	# NG 10-11-10: added clause to exclude NULLs
-# 	$query=<<QUERY
-# 	SELECT type_003 FROM maptable_003
-# 	UNION
-# 	SELECT type_003 FROM maptable_002
-# QUERY
-	$query=<<QUERY
-	SELECT type_003 FROM maptable_003 WHERE type_003 IS NOT NULL
-	UNION
-	SELECT type_003 FROM maptable_002 WHERE type_003 IS NOT NULL
-QUERY
-  ;
-	$view=0;
-      } elsif ($i==3) {
-	$inputs="ConnectDots/maptable_003";
-	# NG 10-11-10: added clause to exclude NULLs
-	# $query="SELECT DISTINCT type_004 FROM maptable_003";
-	$query="SELECT DISTINCT type_004 FROM maptable_003 WHERE type_004 IS NOT NULL";
+    if ($i>=2) {
+      $implicit=1; $query=1;
+      if ($i==3) {
 	$view=1;      
       }}
-
-    report_fail(UNIVERSAL::isa($actual,$class),"$label object $i: class") or return 0;
+    isa_ok_quietly($actual,$class,"$label object $i: class") or return 0;
     report_fail($actual->name eq $name,"$label object $i: name") or return 0;
     report_fail($actual->id eq $id,"$label object $i: id") or return 0;
-    # NG 13-06-12: compare as sets 'cuz perl 5.18 no longer preserves order
-    # report_fail(scrunched_eq($actual->inputs,$inputs),"$label object $i: inputs") or return 0;
-    report_fail(cmp_set_quietly(as_list($actual->inputs),as_list($inputs)),
-		"$label object $i: inputs")
-      or return 0;
-    report_fail(scrunched_eq($actual->namespace,$namespace),"$label object $i: namespace") 
-      or return 0;
-    # NG 13-06-13: for $i==2 (type_003_master), query is UNION over columns from 2 MapTables
-    #              as of perl 5.18, order of MapTables not guaranteed
-    if ($i!=2) {
-      report_fail(scrunched_eq($actual->query,$query),"$label object $i: query") or return 0;
-    } else {
-      my @actual=split(/\s+UNION\s+/,$actual->query);
-      $query=~s/^\s+|\s+$//mg;
-      my @correct=split(/\s+UNION\s+/,$query);
-      report_fail(cmp_set_quietly(\@actual,\@correct),"$label object $i: query") or return 0;
-    }
-    report_fail(as_bool($actual->view)==$view,"$label object $i: view") or return 0;
     report_fail(as_bool($actual->implicit)==$implicit,"$label object $i: implicit") or return 0;
     if ($mature) {
+      report_fail(as_bool($actual->query)==$query,"$label object $i: query") or return 0;
+      report_fail(as_bool($actual->view)==$view,"$label object $i: view") or return 0;
       check_object_basics($actual->babel,'Data::Babel','test',"$label object $i babel");
       check_object_basics($actual->idtype,'Data::Babel::IdType',
 			  "type_$suffix","$label object $i idtype");
@@ -738,29 +892,17 @@ sub check_handcrafted_maptables {
   my $class='Data::Babel::MapTable';
   report_fail(@$actual==$num,"$label: number of elements") or return 0;
   my @actual=sort_objects($actual,$label) or return 0;
+  # NG 13-09-02: DEPRECATED workflow related attributes
   for my $i (0..$#actual) {
     my $actual=$actual[$i];
     my $suffix='00'.($i+1);
     my $suffix1='00'.($i+2);
     my $name="maptable_$suffix";
     my $id="maptable:$name";
-    my $inputs="MainData/table_$suffix";
-    my $query=<<QUERY
-SELECT col_$suffix AS type_$suffix, col_$suffix1 AS type_$suffix1
-FROM   table_$suffix
-QUERY
-      ;
     report_fail(UNIVERSAL::isa($actual,$class),"$label object $i: class") or return 0;
     report_fail($actual->name eq $name,"$label object $i: name") or return 0;
     report_fail($actual->id eq $id,"$label object $i: id") or return 0;
     # NG 13-06-12: compare as sets 'cuz perl 5.18 no longer preserves order
-    # report_fail(scrunched_eq($actual->inputs,$inputs),"$label object $i: inputs") or return 0;
-    report_fail(cmp_set_quietly(as_list($actual->inputs),as_list($inputs)),
-		"$label object $i: inputs")
-      or return 0;
-    report_fail(scrunched_eq($actual->namespace,"ConnectDots"),"$label object $i: namespace") 
-      or return 0;
-    report_fail(scrunched_eq($actual->query,$query),"$label object $i: query") or return 0;
      if ($mature) {
       check_object_basics($actual->babel,'Data::Babel','test',"$label object $i babel");
       check_objects_basics($actual->idtypes,'Data::Babel::IdType',
@@ -880,6 +1022,68 @@ sub _has_history {
 #   ref $idtype and $idtype=$idtype->name;
 #   "_X_$idtype";
 # }
+
+# debugging functions
+
+# print names 
+sub pnames {
+  my $sep=shift;
+  print join($sep,map {$_->name} @_);
+}
+# print graph, typically as sif
+sub pgraph {
+  my($graph,$file,$format)=@_;
+  length($format)? $format=lc($format): ($format='sif');
+  confess "Invalid format $format: must be sif or txt" unless $format=~/sif|txt/;
+  if ($file) {
+    open(OUT,'>',$file) || confess "Cannot create output file $file: $!";
+  } else {
+    *OUT=*STDOUT;
+  }
+  if ($format eq 'sif') {
+    print OUT join("\n",map {my($v0,$v1)=@$_;"$v0 - $v1"} $graph->edges),"\n";
+  } else {
+    print OUT '  ',join("\n  ",map {_edge_str($graph,$_)} _sort_edges($graph->edges)),"\n";
+  }
+  close OUT if $file;
+}
+# # NG 13-10-05 print table dumps to track down FAILs seen by 
+# #              David Cantrell (reports 34101829, 34102877)
+# sub diag_rows {
+#   my($rows)=@_;
+#   my @diag='----------';
+#   for my $row (@$rows) {
+#     # replace undef by NULL
+#     push(@diag,join("\t",map {defined $_? $_: 'NULL'} @$row));
+#   }
+#   push(@diag,'----------');
+#   my $diag=join("\n",@diag);
+#   diag($diag);
+#   return 1;
+# }
+# # NG 13-09-15: print table dumps to track down FAILs seen by 
+# #              David Cantrell (reports 34101829, 34102877)
+# sub diag_table {
+#   my($table,@cols)=@_;
+#   my $cols=@cols? join(',',@cols): '*';
+#   my $sth=$dbh->prepare(qq(SELECT $cols FROM $table)) or goto FAIL;
+#   $sth->execute() or goto FAIL;
+#   my @cols=@{$sth->{NAME}};
+#   my $rows=$sth->fetchall_arrayref() or goto FAIL;
+#   my @diag=("table $table:",join("\t",@cols));
+#   for my $row (@$rows) {
+#     # replace undef by NULL
+#     push(@diag,join("\t",map {defined $_? $_: 'NULL'} @$row));
+#   }
+#   push(@diag,'----------');
+#   my $diag=join("\n",@diag);
+#   diag($diag);
+#   return 1;
+#  FAIL:
+#   fail("dump table $table");
+#   diag("While trying to dump table $table for diagnostic purposes, we got the following DBI error message\n".DBI->errstr);
+#   return 0;
+# }
 1;
 
 package t::FullOuterJoinTable;
